/* Edit: 6/15/2016 Garrison Johnston*/
#include <Servo.h>
#include <math.h>
#include <calibration.h>

const int SERVO[3]={9,6,5}; //Defines Servo Pins

Servo myServo[3]; //creates 3 servo objects

// Defines Robot Geometry 
double rp[3] = {0.031,0.031,0.031};
double rb = 0.16/2;
double L1 = 0.05;
double L2 = 0.05;
double px = 0.0;
double py = 0.0;
double theta = 0.0;
double q1;
double q2;
double q3;
int out_of_reach;

void setup()
{
 //Attaches Servos
  myServo[0].attach(SERVO[0] , ServoMinMicrosecDelay , ServoMaxMicrosecDelay); //specifies min and max pwm width in microseconds
  myServo[1].attach(SERVO[1] , ServoMinMicrosecDelay , ServoMaxMicrosecDelay); //specifies min and max pwm width in microseconds
  myServo[2].attach(SERVO[2] , ServoMinMicrosecDelay , ServoMaxMicrosecDelay); //specifies min and max pwm width in microseconds
  Serial.begin(9600);
  
}

void loop()
{
  double theta_r;

  double xValue = analogRead(A0);
  double yValue = analogRead(A1);
  delay(1);
  px = interpolate(xValue, min_pot, max_pot, -10.0, 10.0);
  py = interpolate(yValue, min_pot, max_pot, -10.0, 10.0);
  
  Serial.println("X Position: "); Serial.print(px); Serial.print(", Y Position: "); Serial.println(py);
  
 //calculates q1, q2, and q3 given px, py, and theta
  InvKinRRR(rb, rp, L1, L2, px, py, (double)theta * pi / 180, &q1, &q2, &q3, &out_of_reach); 
  
  //interpolates pulse width based on calibration 
  m[0]=(int) interpolate(q1+q_home_offset[0]*pi/180, theta_min[0], theta_max[0], m_of_theta_min[0], m_of_theta_max[0]);
  m[1]=(int) interpolate(q2+q_home_offset[1]*pi/180, theta_min[0], theta_max[1], m_of_theta_min[1], m_of_theta_max[1]);
  m[2]=(int) interpolate(q3+q_home_offset[2]*pi/180, theta_min[2], theta_max[2], m_of_theta_min[2], m_of_theta_max[2]);

  //moves servos to desired position
  myServo[0].writeMicroseconds(m[0]);   
  myServo[1].writeMicroseconds(m[1]);          
  myServo[2].writeMicroseconds(m[2]);          
  
  delay(1); // waits for the servo
}

//========================================================================================//
//inverse kinematics function calculates q1, q2, and q3 based on px, py, and theta
void InvKinRRR(double rb, double rp[3], double L1, double L2, double px, double py, double theta, double* q1, double* q2, double* q3, int* out_of_reach) {
  double b1x = rb * cos(-pi / 6);
  double b1y = rb * sin(-pi / 6);
  double b2x = 0;
  double b2y = rb;
  double b3x = rb * cos(7 * pi / 6);
  double b3y = rb * sin(7 * pi / 6);
  double a_angles[3]={-pi/6, pi/2, 7*pi/6};
  double a1x = px + rp[0] * cos(theta + a_angles[0]);
  double a1y = py + rp[0] * sin(theta + a_angles[0]);
  double a2x = px + rp[1] * cos(theta + a_angles[1]);
  double a2y = py + rp[1] * sin(theta + a_angles[1]);
  double a3x = px + rp[2] * cos(theta + a_angles[2]);
  double a3y = py + rp[2] * sin(theta + a_angles[2]);

  double a1b1 = sqrt((a1x - b1x) * (a1x - b1x) + (a1y - b1y) * (a1y - b1y));
  double a2b2 = sqrt((a2x - b2x) * (a2x - b2x) + (a2y - b2y) * (a2y - b2y));
  double a3b3 = sqrt((a3x - b3x) * (a3x - b3x) + (a3y - b3y) * (a3y - b3y));

  double alpha1 = acos((a1b1 * a1b1 +  L1 * L1 - L2 * L2) / (2 * L1 * a1b1));
  double alpha2 = acos((a2b2 * a2b2 +  L1 * L1  - L2 * L2) / (2 * L1 * a2b2));
  double alpha3 = acos((a3b3 * a3b3 +  L1 * L1 - L2 * L2) / (2 * L1 * a3b3));

  double psi1 = atan2(a1y - b1y, a1x - b1x);
  double psi2 = atan2(a2y - b2y, a2x - b2x);
  double psi3 = atan2(a3y - b3y, a3x - b3x);

  psi1 = unwrap(psi1);
  psi2 = unwrap(psi2);
  psi3 = unwrap(psi3);

  *q1 = unwrap(psi1 - alpha1);
  *q2 = unwrap(psi2 - alpha2 - (double)(2 * pi / 3));
  *q3 = unwrap(psi3 - alpha3 - (double)(4 * pi / 3));

  *out_of_reach = 0; //initialize with 0
  if (isnan(*q1) || isnan(*q2) || isnan(*q3)) {
    *out_of_reach = 1;
  }
}

//=============================================================
double unwrap(double theta) {
  const double pi = 3.141592653589793238462643383;
  if (theta < 0) {
    theta = theta + 2 * pi;
  }
  return theta;
}
//=============================================================
//interpolates between two values
double interpolate(double x, double x_min, double x_max, double y_of_x_min, double y_of_x_max) { //interpolates values of m given desired q
  double y;
  y = (double) ((y_of_x_max-y_of_x_min)/(x_max-x_min)*(x-x_min) + y_of_x_min);
  return y;
}
